# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import datetime
from django.utils.timezone import utc
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('django_outlook', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='mailhandler',
            options={'ordering': ('-modified', '-created'), 'get_latest_by': 'modified'},
        ),
        migrations.AddField(
            model_name='mailhandler',
            name='created',
            field=django_extensions.db.fields.CreationDateTimeField(default=datetime.datetime(2018, 6, 28, 4, 40, 38, 895000, tzinfo=utc), verbose_name='created', auto_now_add=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='mailhandler',
            name='is_enabled',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='mailhandler',
            name='modified',
            field=django_extensions.db.fields.ModificationDateTimeField(default=datetime.datetime(2018, 6, 28, 4, 40, 44, 748000, tzinfo=utc), verbose_name='modified', auto_now=True),
            preserve_default=False,
        ),
    ]



from django.views.generic import TemplateView, RedirectView
from django_outlook.o365_utils.adv_connection import OutlookConnection
from django_outlook.o365_utils.login_token_storage import LoginTokenStorage
from django_outlook.o365_utils.token_storage import TokenStorage
from djangoautoconf.django_utils import retrieve_param
from djangoautoconf.local_key_manager import get_local_key


class O365AuthRedirectView(RedirectView):
    permanent = False  # Not always redirect to the same page

    def get_redirect_url(self, *args, **kwargs):
        # article = get_object_or_404(Article, pk=kwargs['pk'])
        # article.update_counter()
        # return super().get_redirect_url(*args, **kwargs)
        token_storage = TokenStorage(self.request.user)
        c = OutlookConnection(
            get_local_key("o365_app_settings.o365_app_client_id"),
            get_local_key("o365_app_settings.o365_app_secret"),
            token_storage,
        )
        auth_url = c.get_auth_url()
        return auth_url


class OutlookAuthResultView(TemplateView):
    template_name = 'django_outlook/key_got.html'

    def get_context_data(self, **kwargs):
        # return super(OutlookAuthResultView, self).get_context_data(**kwargs)
        # param = retrieve_param(self.request)
        token_storage = TokenStorage(self.request.user)
        c = OutlookConnection(
            get_local_key("o365_app_settings.o365_app_client_id"),
            get_local_key("o365_app_settings.o365_app_secret"),
            token_storage,
        )
        token_url = "%s/?%s" % ("https://localhost", self.request.META['QUERY_STRING'])

        res = c.update_extra_data(token_url)
        return res


class O365LoginRedirectView(RedirectView):
    permanent = False  # Not always redirect to the same page

    def get_redirect_url(self, *args, **kwargs):
        # article = get_object_or_404(Article, pk=kwargs['pk'])
        # article.update_counter()
        # return super().get_redirect_url(*args, **kwargs)
        token_storage = LoginTokenStorage(self.request.user)
        c = OutlookConnection(
            get_local_key("o365_login_app_settings.o365_app_client_id"),
            get_local_key("o365_login_app_settings.o365_app_secret"),
            token_storage,
            redirect_url='https://%s/django_outlook/login_result/' % self.request.get_host()
        )
        auth_url = c.get_auth_url()
        return auth_url


class OutlookLoginResultView(TemplateView):
    template_name = 'django_outlook/key_got.html'

    def get_context_data(self, **kwargs):
        # return super(OutlookLoginResultView, self).get_context_data(**kwargs)
        # param = retrieve_param(self.request)
        token_storage = LoginTokenStorage(self.request.user)
        c = OutlookConnection(
            get_local_key("o365_login_app_settings.o365_app_client_id"),
            get_local_key("o365_login_app_settings.o365_app_secret"),
            token_storage,
            redirect_url='https://%s/django_outlook/login_result/' % self.request.get_host()
        )
        token_url = "%s/?%s" % ("https://localhost", self.request.META['QUERY_STRING'])
        param = retrieve_param(self.request)
        res = c.update_extra_data(token_url, param["state"])
        return res



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields.json


class Migration(migrations.Migration):

    dependencies = [
        ('default', '0003_alter_email_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='MailHandler',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('action', models.CharField(max_length=1024)),
                ('parameters', django_extensions.db.fields.json.JSONField()),
                ('auth', models.ForeignKey(to='default.UserSocialAuth')),
            ],
        ),
    ]



import logging
import os

from oauthlib.oauth2 import TokenExpiredError
from requests_oauthlib import OAuth2Session

from O365.connection import MicroDict
from djangoautoconf.local_key_manager import get_local_key

log = logging.getLogger(__name__)


# Ref: O365
class OutlookConnection(object):
    _oauth2_authorize_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/authorize'
    _oauth2_token_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/token'
    api_version = '2.0'

    def __init__(self, client_id, client_secret, token_storage, redirect_url="https://outlook.office365.com/owa/"):
        super(OutlookConnection, self).__init__()
        self.redirect_url = redirect_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.client_id = self.client_id
        self.client_secret = self.client_secret
        self.token_storage = token_storage
        self.proxy_dict = None
        self.oauth = None
        # self.auth = None

        # Proxy call is required only if you are behind proxy
        self.set_proxy(url=get_local_key("proxy_setting.http_proxy_host"),
                       port=8080,
                       username=get_local_key("laptop_account.username"),
                       password=get_local_key("laptop_account.password")
                       )

    def set_proxy(self, url, port, username, password):
        """ Connect to Office 365 though the specified proxy

        :param url: url of the proxy server
        :param port: port to connect to proxy server
        :param username: username for authentication in the proxy server
        :param password: password for the specified username
        """

        self.proxy_dict = {
            "http": "http://{}:{}@{}:{}".format(username, password, url, port),
            "https": "https://{}:{}@{}:{}".format(username, password, url,
                                                  port),
        }

    def load_token(self):
        self.oauth = OAuth2Session(client_id=self.client_id,
                                   token=self.token_storage.get_token())

    def get_auth_url(self):
        self._set_oauth_session()
        auth_url, state = self.oauth.authorization_url(
            url=self._oauth2_authorize_url,
            access_type='offline')
        # username = user.username
        self.token_storage.save_state(state)
        return auth_url

    def update_extra_data(self, auth_resp, state):
        try:
            if state is None:
                state = self.token_storage.get_stored_state()
            token = self._get_token(auth_resp, state)
            me_dict = self.get_me()
            self.token_storage.save_first_token(token, me_dict)
            res = {"json_key": str(token)}
        except Exception as e:
            res = {"json_key": str(e.message)}
        return res

    def _get_token(self, auth_resp, state):
        self._set_oauth_session(state)
        self.oauth.authorization_url(
            url=self._oauth2_authorize_url,
            access_type='offline')
        os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = 'Y'
        token = self.oauth.fetch_token(token_url=self._oauth2_token_url,
                                       authorization_response=auth_resp, client_secret=self.client_secret)

        return token

    def _set_oauth_session(self, state=None):
        self.oauth = OAuth2Session(client_id=self.client_id,
                                   redirect_uri=self.redirect_url,
                                   scope=[
                                       'https://graph.microsoft.com/Mail.ReadWrite',
                                       'https://graph.microsoft.com/Mail.Send',
                                       'https://graph.microsoft.com/Mail.Read.Shared',
                                       'https://graph.microsoft.com/Mail.ReadWrite.Shared',
                                       'https://graph.microsoft.com/User.Read',
                                       # 'https://outlook.office.com/mail.readwrite',
                                       'offline_access',
                                       'https://graph.microsoft.com/mailboxsettings.readwrite'
                                   ],
                                   state=state
                                   )

    def get_response(self, request_url, **kwargs):
        """ Fetches the response for specified url and arguments, adding the auth and proxy information to the url

        :param request_url: url to request
        :param kwargs: any keyword arguments to pass to the requests api
        :return: response object
        """
        response_json = self.get_common_response(request_url, **kwargs)

        if 'value' not in response_json:
            raise RuntimeError('Something went wrong, received an unexpected result \n{}'.format(response_json))

        response_values = [MicroDict(x) for x in response_json['value']]
        return response_values

    def get_common_response(self, request_url, **kwargs):
        con_params = {}
        if self.proxy_dict:
            con_params['proxies'] = self.proxy_dict
        con_params.update(kwargs)
        log.info('Requesting URL: {}'.format(request_url))
        try:
            response = self.oauth.get(request_url, **con_params)
        except TokenExpiredError:
            self.refresh_token()

            response = self.oauth.get(request_url, **con_params)
        log.info('Received response from URL {}'.format(response.url))
        response_json = response.json()
        return response_json

    def post_common_response(self, request_url, **kwargs):
        con_params = {}
        if self.proxy_dict:
            con_params['proxies'] = self.proxy_dict
        con_params.update(kwargs)
        log.info('Requesting URL: {}'.format(request_url))
        try:
            response = self.oauth.post(request_url, **con_params)
        except TokenExpiredError:
            self.refresh_token()

            response = self.oauth.post(request_url, **con_params)
        log.info('Received response from URL {}'.format(response.url))
        response_json = response.json()
        return response_json

    def refresh_token(self):
        log.info('Token is expired, fetching a new token')
        token = self.oauth.refresh_token(self._oauth2_token_url,
                                         client_id=self.client_id,
                                         client_secret=self.client_secret)
        log.info('New token fetched')
        self.token_storage.save_token(token)

    def get_me(self):
        """
        :return: Current auto reply setting
        """
        auto_reply_setting_url = 'https://graph.microsoft.com/v1.0/me'

        response = self.get_common_response(auto_reply_setting_url,
                                            # verify=self.verify,
                                            # params={'$top': 100}
                                            )

        return response



import string
from random import choice, randint

from django.contrib.auth.models import User
from social.apps.django_app.default.models import UserSocialAuth

from django_outlook.o365_utils.token_storage import EmailForUserToAccessDoesNotExist


class LoginTokenStorage(object):
    """
    Store token in UserSocialAuth
    Need to keep the following info:
    1. user whose account is bind to this account
    2. user account (email) of this social auth
    3. service name
    4. role of this social auth

    uid: social account owner email
    provider: o365-outlook (if it is ongoing: o365-ongoing)
    user: current user
    """
    PROVIDER_FOR_ONGOING_AUTH = "o365-ongoing"
    PROVIDER = "o365-outlook"

    def __init__(self, current_user=None, mail_of_user_grant_the_token=None):
        super(LoginTokenStorage, self).__init__()
        self.mail_of_user_grant_the_token = mail_of_user_grant_the_token

    def get_token(self):
        auth = self._get_stored_social_auth()
        return auth.extra_data

    def save_first_token(self, token, user_info):
        characters = string.ascii_letters + string.punctuation + string.digits
        password = "".join(choice(characters) for x in range(randint(8, 16)))
        user = self._get_user_from_email(user_info["mail"])
        user.password = password
        user.save()
        o, is_social_auth_created = UserSocialAuth.objects.get_or_create(
            user=user,
            provider=self.PROVIDER,
            uid=user_info["id"],
        )
        o.extra_data = token
        o.save()

    def save_token(self, token):
        o = self._get_stored_social_auth()
        o.extra_data = token
        o.save()

    def save_state(self, state):
        pass

    # noinspection PyMethodMayBeStatic
    def _get_user_from_email(self, user_email):
        user, is_user_created = User.objects.get_or_create(username=user_email,
                                                           email=user_email)
        return user

    def _get_stored_social_auth(self):
        if self.mail_of_user_grant_the_token is None:
            raise EmailForUserToAccessDoesNotExist
        user = self._get_user_from_email(self.mail_of_user_grant_the_token)
        auth = UserSocialAuth.objects.get(
            user=user,
            provider=self.PROVIDER,
        )
        return auth



# noinspection PyMethodMayBeStatic
import logging


class MailToHandleBase(object):
    logger = logging.getLogger("MailToHandleBase")

    def is_valid(self):
        return False

    def get_sender(self):
        return []

    def get_entry_id(self):
        return None

    def get_recipients(self):
        return []

    def get_categories(self):
        return []

    def get_subject(self):
        return None

    def get_conversation_id(self):
        return None

    def get_receive_datetime(self):
        return None

    def get_mail_specific_link(self):
        return self.get_one_note_link()

    def get_one_note_link(self):
        return 'onenote:outlook?folder=Mail&entryid=' + str(self.get_entry_id())

    def dump_mail(self):
        # properties can be found at https://msdn.microsoft.com/en-us/library/office/dn320330.aspx
        email_info = unicode(self.get_subject()), unicode(self.get_entry_id()), \
                     unicode(self.get_sender()), \
                     "conversationID:", unicode(self.get_conversation_id()), \
                     "category:", unicode(self.get_categories())
        self.logger.info(email_info)

    def get_body(self):
        return None



from social.apps.django_app.default.models import UserSocialAuth


class EmailForUserToAccessDoesNotExist(Exception):
    pass


class TokenStorage(object):
    """
    Store token in UserSocialAuth
    Need to keep the following info:
    1. user whose account is bind to this account
    2. user account (email) of this social auth
    3. service name
    4. role of this social auth

    uid: social account owner email
    provider: o365-outlook (if it is ongoing: o365-ongoing)
    user: current user
    """
    PROVIDER_FOR_ONGOING_AUTH = "o365-ongoing"
    PROVIDER = "o365-outlook"

    def __init__(self, current_user, mail_of_user_grant_the_token=None):
        super(TokenStorage, self).__init__()
        self.current_user = current_user
        self.mail_of_user_grant_the_token = mail_of_user_grant_the_token

    def get_token(self):
        if self.mail_of_user_grant_the_token is None:
            raise EmailForUserToAccessDoesNotExist
        auth = self.get_auth_obj()
        return auth.extra_data

    def save_first_token(self, token, user_info):
        mail_of_user_grant_the_token = user_info["mail"]
        o = self._get_ongoing_social_auth()
        if self.is_stored_social_auth_exists(mail_of_user_grant_the_token):
            o.delete()
            o = self.get_stored_social_auth(mail_of_user_grant_the_token)
        else:
            o.uid = mail_of_user_grant_the_token
            o.provider = self.PROVIDER
        o.extra_data = token
        o.save()

    def get_stored_social_auth(self, uid):
        return UserSocialAuth.objects.get(
            user=self.current_user,
            provider=self.PROVIDER,
            uid=uid,
        )

    def save_token(self, token):
        if self.mail_of_user_grant_the_token is None:
            raise EmailForUserToAccessDoesNotExist
        o = self.get_auth_obj()
        o.extra_data = token
        o.save()

    def get_auth_obj(self):
        return self.get_stored_social_auth(self.get_uid())

    def _get_ongoing_social_auth(self):
        return UserSocialAuth.objects.get(
            user=self.current_user,
            provider=self.PROVIDER_FOR_ONGOING_AUTH,
            uid=self.get_uid(),
        )

    def get_stored_state(self):
        social_auth = self._get_ongoing_social_auth()
        state = social_auth.extra_data["state"]
        return state

    def save_state(self, state):
        o, is_created = UserSocialAuth.objects.get_or_create(
            user=self.current_user,
            provider=self.PROVIDER_FOR_ONGOING_AUTH,
            uid=(self.get_uid()),
        )
        o.extra_data = {"state": state}
        o.save()

    def get_uid(self):
        if self.mail_of_user_grant_the_token is None:
            uid = "unknown"
        else:
            uid = self.mail_of_user_grant_the_token
        return uid

    def set_mail_of_user_grant_the_token(self, mail):
        self.mail_of_user_grant_the_token = mail

    def is_stored_social_auth_exists(self, uid):
        return UserSocialAuth.objects.filter(
            user=self.current_user,
            provider=self.PROVIDER,
            uid=uid,
        ).exists()



from O365 import FluentInbox, Connection
from django_outlook.o365_utils.adv_connection import OutlookConnection


class AdvO365Mailbox(FluentInbox):
    adv_urls = {
        'auto_reply_setting': {
            '1.0': 'https://outlook.office365.com/api/v1.0/me/MailboxSettings/AutomaticRepliesSetting',
            '2.0': 'https://graph.microsoft.com/v1.0/me/MailboxSettings/AutomaticRepliesSetting',
        },
        'me': {
            '1.0': 'https://outlook.office365.com/api/v1.0/me',
            '2.0': 'https://graph.microsoft.com/v1.0/me',
        },
    }

    def get_auto_reply_setting(self):
        """
        :return: Current auto reply setting
        """
        auto_reply_setting_url = AdvO365Mailbox._get_url_v2('auto_reply_setting')

        response = OutlookConnection.get_common_response(auto_reply_setting_url,
                                                         # verify=self.verify,
                                                         # params={'$top': 100}
                                                         )

        return response

    @staticmethod
    def _get_url_v2(key):
        """ Fetches the url for specified key as per the connection version configured

        :param key: the key for which url is required
        :return: URL to use for requests
        """
        return AdvO365Mailbox.adv_urls[key][Connection().api_version]



from django_outlook.management.mail_to_handle_base import MailToHandleBase


class OutlookMail(MailToHandleBase):
    pre_check_attr_list = ["EntryID", "To"]

    def __init__(self, mail):
        super(OutlookMail, self).__init__()
        self.outlook_mail = mail
        self._is_valid = True
        try:
            for attr in self.pre_check_attr_list:
                getattr(mail, attr)
        except AttributeError:
            self._is_valid = False
        except Exception, e:
            self._is_valid = False

    def is_valid(self):
        return self._is_valid

    def get_categories(self):
        try:
            return unicode(self.outlook_mail.Categories).split(",")
        except AttributeError:
            return []

    def get_sender(self):
        try:
            sender = self.outlook_mail.SenderName
        except:
            sender = "No sender"
        return sender

    def get_subject(self):
        return self.outlook_mail.Subject

    def get_entry_id(self):
        return self.outlook_mail.EntryID

    def get_recipients(self):
        return self.outlook_mail.To

    def get_conversation_id(self):
        return self.outlook_mail.ConversationId

    def get_receive_datetime(self):
        return self.outlook_mail.receivedTime

    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except:
            pass
        return getattr(self.outlook_mail, name)

    def dump_mail(self):
        # properties can be found at https://msdn.microsoft.com/en-us/library/office/dn320330.aspx
        email_info = unicode(self.get_subject()), unicode(self.get_entry_id()), \
                     unicode(self.get_sender()), \
                     unicode(self.outlook_mail.BodyFormat), "session: ", unicode(self.outlook_mail.Session), \
                     "conversationID:", unicode(self.get_conversation_id()), \
                     "c index:", unicode(self.outlook_mail.ConversationIndex), \
                     "topic:", unicode(self.outlook_mail.ConversationTopic), \
                     "category:", unicode(self.get_categories()), \
                     "unread:", unicode(self.outlook_mail.Unread), \
                     "receivedTime", unicode(self.get_receive_datetime())
        self.logger.info(email_info)

    def get_body(self):
        return self.outlook_mail.body

    def move_to_folder(self, target_folder):
        self.outlook_mail.move(target_folder)




import base64

from django_outlook.management.mail_to_handle_base import MailToHandleBase
import urllib
from bs4 import BeautifulSoup


class O365Mail(MailToHandleBase):

    def __init__(self, raw_mail):
        super(O365Mail, self).__init__()
        self.raw_mail = raw_mail
    #
    # def get_categories(self):
    #     return self.raw_mail["Categories"]

    def __getattr__(self, name):
        try:
            return self.raw_mail.json[name]
        except:
            pass
        return getattr(self.raw_mail, name)

    def get_entry_id(self):
        # Ref: https://stackoverflow.com/questions/2941995/python-ignore-incorrect-padding-error-when-base64-decoding
        return base64.b64decode(str(self.raw_mail.json["id"]), '-_').encode("hex")[-140:]

    def get_conversation_id(self):
        # Ref https://stackoverflow.com/questions/2941995/python-ignore-incorrect-padding-error-when-base64-decoding
        return base64.b64decode(str(self.raw_mail.json["conversationId"]), '-_').encode("hex")[-32:]

    def get_categories(self):
        return self.raw_mail.json["categories"]

    def get_subject(self):
        return self.raw_mail.json["subject"]

    def is_valid(self):
        return True

    def get_sender(self):
        return self.raw_mail.json["sender"]["emailAddress"]["address"]

    def get_receive_datetime(self):
        return self.raw_mail.json["receivedDateTime"]

    def get_recipients(self):
        res = []
        for to in self.raw_mail.json["toRecipients"]:
            res.append(to["emailAddress"]["address"])
        return res

    def get_mail_specific_link(self):
        return "%s\n%s" % (self.get_one_note_link(), self.raw_mail.json["webLink"])

    def get_body(self):
        return self._html_to_text(self.raw_mail.json["body"]["content"])

    def _html_to_text(self, html):
        # https://stackoverflow.com/questions/328356/extracting-text-from-html-file-using-python
        soup = BeautifulSoup(html, "lxml")

        # kill all script and style elements
        for script in soup(["script", "style"]):
            script.extract()    # rip it out

        # get text
        text = soup.get_text()

        # break into lines and remove leading and trailing space on each
        lines = (line.strip() for line in text.splitlines())
        # break multi-headlines into a line each
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        # drop blank lines
        text = '\n'.join(chunk for chunk in chunks if chunk)

        return text

    def move_to_folder(self, target_folder):
        self.raw_mail.moveToFolder(target_folder)



from django.contrib.auth.models import User
from django.db import models
from django_extensions.db.fields.json import JSONField
from django_extensions.db.models import TimeStampedModel
from social.apps.django_app.default.models import AbstractUserSocialAuth, UserSocialAuth


# class MicrosoftOauth(models.Model):
#     token_type = models.CharField(max_length=128)
#     refresh_token = models.TextField()
#     scope = models.TextField()
#     access_token = models.TextField()
#     expires_in = models.IntegerField()
#     ext_expires_in = models.IntegerField()
#     expires_at = models.FloatField()
#     user = models.ForeignKey(User)


class MailHandler(TimeStampedModel):
    auth = models.ForeignKey(UserSocialAuth)
    action = models.CharField(max_length=1024)
    parameters = JSONField(null=True, blank=True)
    is_enabled = models.BooleanField(default=True)





class O365UrlGenerator(object):

    def __init__(self, target_mail):
        super(O365UrlGenerator, self).__init__()
        self.target_mail = target_mail

    def get_inbox_url(self):
        """
        Ref: https://stackoverflow.com/questions/40813610/microsoft-graph-api-net-not-able-to-read-shared-mail
        Ref: o365
        :return:
        """
        return 'https://graph.microsoft.com/v1.0/users/{user_id}/messages'.format(user_id=self.target_mail)

    def get_folders_url(self):
        return 'https://graph.microsoft.com/v1.0/users/{user_id}/MailFolders'.format(user_id=self.target_mail)



from django.conf.urls import patterns, url
from django.contrib.auth.decorators import login_required

import models
from django_outlook.views import OutlookLoginResultView, O365AuthRedirectView, O365LoginRedirectView, \
    OutlookAuthResultView
from djangoautoconf.model_utils.url_for_models import add_all_urls


urlpatterns = [
    url(r'login_result/', OutlookLoginResultView.as_view(), name="o365_login_view"),
    url(r'login/', O365LoginRedirectView.as_view(), name="o365_login_redirect_view"),
    url(r'result/', login_required(OutlookAuthResultView.as_view()), name="o365_token_update_view"),
    url(r'', login_required(O365AuthRedirectView.as_view()), name="o365_auth_redirect_view"),
    ]

add_all_urls(urlpatterns, models)
